require "bezier"
require "bspline"

mode = { bezier = {}, bspline = {}}

-- == -- == -- == -- == -- == -- == -- == -- == -- == -- == -- == --
function mode.bezier.load()
	draw_control_polygon = true
	smoothness = 3
	b = Bezier.new(points)
	b_smooth = b:polygon(smoothness)
end

function mode.bezier.draw()
	if b then
		love.graphics.setColor(0,0,100,255)
		love.graphics.line(unpack(b_smooth))

		local p = b:eval(t)
		love.graphics.circle('line', p.x, p.y, 5)
	end

	if bdrag then
		love.graphics.setColor(0,0,100,50)
		bdrag:draw(2)
		love.graphics.setColor(0,0,100,30)
		love.graphics.line(unpack(bdrag:controlPolygon()))
	end

	if draw_control_polygon then
		if b then
			love.graphics.setColor(0,0,100,100)
			love.graphics.line(unpack(b:controlPolygon()))
		end
		love.graphics.setColor(0,100,0,255)
		for i = 1,#points/2 do
			local x,y = points[2*i-1], points[2*i]
			love.graphics.circle('fill', x, y, 4, 4)
			love.graphics.print(tostring(i), x, y - 10, 0, .7)
		end
	end

	love.graphics.setColor(0,0,0,255)
	love.graphics.print(string.format("Bzier-curve (degree: %d, control points: %d)",
						b and b.degree or 0, b and #b.controlPoints or #points/2), 10, 20)
	love.graphics.setColor(0,0,0,160)
	love.graphics.print(string.format("FPS: %d, t = %.2f Smoothness: %d",
						love.timer.getFPS(), t, smoothness), 10, 40)
	love.graphics.printf([[[mouse] place control point
[rmouse] drag
[c] clear control points
[d] degree up
[s] subdivide
[g] derivation curve
[t] toggle control polygon
[o] increase smoothness
[p] decrese smoothness
[m] toggle mode
]], 610, 20, 180)
end

function mode.bezier.keyreleased(key)
	if key == 'c' then
		points = {}
		b = nil
	elseif key == 'd' then
		b:degreeUp()
		b_smooth = b:polygon(smoothness)
		points = b:controlPolygon()
	elseif key == 's' then
		b:subdivide()
		b_smooth = b:polygon(smoothness)
		points = b:controlPolygon()
	elseif key == 't' then
		draw_control_polygon = not draw_control_polygon
	elseif key == 'g' then
		b = b:derivation()
		b_smooth = b:polygon(smoothness)
		points = b:controlPolygon()
	elseif key == 'o' then
		smoothness = math.min(smoothness + 1, 7)
		b_smooth = b:polygon(smoothness)
	elseif key == 'p' then
		smoothness = math.max(smoothness - 1, 0)
		b_smooth = b:polygon(smoothness)
	end
end

function mode.bezier.update(dt)
	t = t + dt * .25
	t = t % 1

	if startdrag then
		bdrag = b:clone()
		bdrag:translate((vector.new(love.mouse.getPosition())-startdrag):unpack())
	end
end

function mode.bezier.mousereleased(x,y, button)
	if button == 'r' and startdrag then
		b:translate((vector.new(x,y)-startdrag):unpack())
		startdrag = nil
		bdrag = nil

		b_smooth = b:polygon(smoothness)
		points = b:controlPolygon()
	end
end

function mode.bezier.mousepressed(x,y, button)
	if button == 'r' then
		startdrag = vector.new(x,y)
		return
	end
	points[#points+1] = x
	points[#points+1] = y
	if #points > 2 then
		b = Bezier.new(points)
		b_smooth = b:polygon(smoothness)
	end
end

-- == -- == -- == -- == -- == -- == -- == -- == -- == -- == -- == --

function mode.bspline.load()
	smoothness = 4
	draw_control_polygon = true
	s = BSpline.new(points)
	s_smooth = s:clone():polygon(smoothness)
end

function mode.bspline.draw()
	if s then
		love.graphics.setColor(0,0,100,255)
		love.graphics.line(unpack(s_smooth))

		local p = s:eval(t)
		love.graphics.circle('line', p.x, p.y, 5)
	end

	if sdrag then
		love.graphics.setColor(0,0,100,50)
		sdrag:draw(2)
		love.graphics.setColor(0,0,100,30)
		love.graphics.line(unpack(sdrag:controlPolygon()))
	end

	if draw_control_polygon then
		if s then
			love.graphics.setColor(0,0,100,100)
			love.graphics.line(unpack(s:controlPolygon()))
		end
		love.graphics.setColor(0,100,0,255)
		for i = 1,#points/2 do
			local x,y = points[2*i-1], points[2*i]
			love.graphics.circle('fill', x, y, 4, 4)
			love.graphics.print(tostring(i), x, y - 10, 0, .7)
		end
	end

	love.graphics.setColor(0,0,0,255)
	love.graphics.print(string.format("BSpline (degree: %d, control points: %d, knots: %d)",
						s and s.degree or 0, s and #s.controlpoints or #points/2, 
						s and #s.knots or 0), 10, 20)
	love.graphics.setColor(0,0,0,160)
	love.graphics.print(string.format("FPS: %d, t = %.2f, Smoothness: %d",
						love.timer.getFPS(), t, smoothness), 10, 40)
	love.graphics.printf([[[mouse] place control point
[rmouse] drag
[c] clear control points
[s] subdivide
[t] toggle control polygon
[o] increase smoothness
[p] decrese smoothness
[m] toggle mode
]], 610, 20, 180)
end

function mode.bspline.keyreleased(key)
	if key == 'c' then
		s = nil
		s_smooth = nil
		points = {}
	elseif key == 's' then
		s:subdivide()
		s_smooth = s:polygon(smoothness)
		points = s:controlPolygon()
	elseif key == 't' then
		draw_control_polygon = not draw_control_polygon
	elseif key == 'o' then
		smoothness = math.min(smoothness+1, 9)
		s_smooth = s:polygon(smoothness)
	elseif key == 'p' then
		smoothness = math.max(smoothness-1, 0)
		s_smooth = s:polygon(smoothness)
	elseif key == 'i' then
		s = BSpline.interpolate(points)
		points = s:controlPolygon()
		s_smooth = s:polygon(smoothness)
	end
end

function mode.bspline.update(dt)
	if s then
		t = t + dt * .25
		t = t % (s:knotMax() + s:knotMin())
		if t < s:knotMin() then
			t = s:knotMin()
		end
	end

	if startdrag then
		sdrag = s:clone()
		sdrag:translate((vector.new(love.mouse.getPosition())-startdrag):unpack())
	end
end

function mode.bspline.mousereleased(x,y, button)
	if button == 'r' and startdrag then
		s:translate((vector.new(x,y)-startdrag):unpack())
		startdrag = nil
		sdrag = nil

		s_smooth = s:polygon(smoothness)
		points = s:controlPolygon()
	end
end

function mode.bspline.mousepressed(x,y, button)
	if button == 'r' then
		startdrag = vector.new(x,y)
		return
	end

	points[#points+1] = x
	points[#points+1] = y
	if #points >= 4 then
		s = BSpline.new(points)
		s_smooth = s:polygon(smoothness)
	end
end

-- == -- == -- == -- == -- == -- == -- == -- == -- == -- == -- == --

function love.load()
	love.graphics.setBackgroundColor(255,255,255)
	love.graphics.setFont(12)
	points = {10,400, 100,100, 400,90, 490,380, 530,200, 700,500, 200,300}
	t = 0
	currentMode = mode.bspline
	currentMode.load()
end

function love.draw()
	currentMode.draw()
end

local nextmode = {
	[mode.bspline] = mode.bezier,
	[mode.bezier]  = mode.bspline,
}
function love.keyreleased(key)
	if key == 'm' then
		currentMode = nextmode[currentMode]
		currentMode.load()
	else
		currentMode.keyreleased(key)
	end
end

function love.update(dt)
	currentMode.update(dt)
end

function love.mousereleased(x,y, button)
	currentMode.mousereleased(x,y,button)
end

function love.mousepressed(x,y, button)
	currentMode.mousepressed(x,y,button)
end
